#ifndef TYPED_LIST_H
#define TYPED_LIST_H

#include "c_cpp_aliases/aliases.h"
#include <assert.h>
#include <string.h>

#define BASE_LIST_CAPACITY 1024

#define CONCAT(A, B, C) A##B##C
#define LIST_TYPE_NAME(T) CONCAT(list_, T, _t)

#define MAKE_LIST_TYPE(T)                                                      \
  typedef struct {                                                             \
    T *items;                                                                  \
    u64 capacity;                                                              \
    u64 count;                                                                 \
  } LIST_TYPE_NAME(T)

#define list_create(T)                                                         \
  (LIST_TYPE_NAME(T) *)_create_list(sizeof(T), BASE_LIST_CAPACITY)
#define list_create_with_capacity(T, CAPACITY)                                 \
  (LIST_TYPE_NAME(T) *)_create_list(sizeof(T), CAPACITY)

#define list_destroy(T, LP)                                                    \
  do {                                                                         \
    munmap(LP->items, sizeof(T) * LP->capacity);                               \
    LP->items = NULL;                                                          \
    LP->count = 0;                                                             \
    LP->capacity = 0;                                                          \
    munmap(LP, sizeof(LIST_TYPE_NAME(T)));                                     \
  } while (0)

#define _increase_list_capacity(T, LP, CAP)                                    \
  do {                                                                         \
    u64 new_capacity = LP->capacity * 2;                                       \
    T *tmp = _alloc_for_list(sizeof(T) * new_capacity);                        \
    assert(tmp != NULL && "Failed to increase capacity");                      \
                                                                               \
    memcpy(tmp, LP->items, sizeof(T) * LP->count);                             \
                                                                               \
    i32 deallocated = munmap(LP->items, sizeof(T) * LP->capacity);             \
    assert(deallocated == 0 && "Failed to deallocate old memory");             \
                                                                               \
    LP->capacity = new_capacity;                                               \
    LP->items = tmp;                                                           \
  } while (0)

#define list_append(T, LP, ITEM)                                               \
  do {                                                                         \
    if (LP->count + 1 >= LP->capacity) {                                       \
      _increase_list_capacity(T, LP, LP->capacity * 2);                        \
    }                                                                          \
                                                                               \
    LP->items[(LP->count)++] = ITEM;                                           \
  } while (0)

#define list_pop(LP) (LP->count -= 1)

#define list_get(LP, IDX) LP->items[IDX]

#define list_merge(T, DST, LP1, LP2)                                           \
  do {                                                                         \
    u64 new_count = LP1->count + LP2->count;                                   \
    u64 capacity =                                                             \
        new_count < BASE_LIST_CAPACITY ? BASE_LIST_CAPACITY : new_count * 2;   \
                                                                               \
    DST = (LIST_TYPE_NAME(T) *)_create_list(new_count, capacity);              \
    assert(DST != NULL && "Failed to allocate new list");                      \
                                                                               \
    u64 lp1_copy_size = sizeof(T) * LP1->count;                                \
    u64 lp2_copy_size = sizeof(T) * LP2->count;                                \
    memcpy(DST->items, LP1->items, lp1_copy_size);                             \
    T *dst = &(DST->items[LP1->count]);                                        \
    memcpy(dst, LP2->items, lp2_copy_size);                                    \
    DST->capacity = capacity;                                                  \
    DST->count = new_count;                                                    \
  } while (0)

MAKE_LIST_TYPE(void);

list_void_t *_create_list(u64 size, u64 count);
void *_alloc_for_list(u64 size);

#endif // !TYPED_LIST_H
